use std::{
    io,
    net::{TcpStream, ToSocketAddrs},
    sync::Arc,
    time::Duration,
};

use rustls::{ClientConfig, ClientConnection, StreamOwned};

pub use self::{error::Error, verifier::Verifier};
use {
    crate::prelude::{CertificateStore, Request},
    std::io::{Read, Write},
};

mod error;
mod verifier;

#[derive(Debug)]
/// Sends a piece of mail from the sending server to the receiving server
pub struct Sender<S>
where
    S: CertificateStore,
{
    /// The full message text to be sent
    pub request: Request,
    /// A [CertificateStore] for servers known to us
    pub store: S,
    /// A [CertificateStore] for mailboxes which exist on this system
    pub client_store: S,
}

impl<S> Sender<S>
where
    S: CertificateStore + 'static,
{
    pub fn new(request_str: &str, store: S, client_store: S) -> Result<Self, Error> {
        let request: Request = request_str.parse()?;
        Ok(Self {
            request,
            store,
            client_store,
        })
    }

    fn host_string(&self) -> String {
        self.request.sender.host.to_string()
    }

    pub fn send(self) -> Result<Vec<u8>, Error> {
        let dnsname = self
            .host_string()
            .as_str()
            .try_into()
            .map_err(|_| Error::DnsError)?;
        let mut it = self.request.sender.host.to_socket_addrs()?;
        let client_cert = self
            .client_store
            .get_certificate(&self.request.sender.to_string());
        let verifier = Arc::new(Verifier::new(self.store));
        let Some(socket_addrs) = it.next() else {
            return Err(io::Error::new(io::ErrorKind::Other, "no data retrieved").into());
        };
        let tcp_stream = TcpStream::connect_timeout(&socket_addrs, Duration::new(10, 0))?;
        let cfg = ClientConfig::builder()
            .with_safe_defaults()
            .with_custom_certificate_verifier(verifier);
        let cfg = match client_cert {
            None => cfg.with_no_client_auth(),
            Some(_) => cfg.with_no_client_auth(), // todo: cfg.with_single_cert(cert_chain, key_der)
        };
        let client = ClientConnection::new(Arc::new(cfg), dnsname)?;
        let mut stream = StreamOwned::new(client, tcp_stream);
        stream.write_all(self.request.to_string().as_bytes())?;
        let mut buf = vec![];
        stream.read_to_end(&mut buf)?;
        stream.conn.send_close_notify();
        drop(stream);
        Ok(buf)
    }
}
